import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import random
import os

import logging
logging.getLogger('tensorflow').disabled = True

# title Dataset Constants
NUM_FREEFORM_TRAIN_TFRECORDS = 100
NUM_FREEFORM_TEST_TFRECORDS = 10

FREEFORM_FEATURES = dict(
     image=tf.io.FixedLenFeature(dtype=tf.string, shape=()),                                                                                                                                             
     mask=tf.io.FixedLenFeature(dtype=tf.string, shape=()),                                                                                                                                              
     camera_pose=tf.io.FixedLenFeature(dtype=tf.float32, shape=(15, 6)),                                                                                                                              
)

# title Dataset Utilities       
                                                                                                                                                                
def parse_freeform_row(row):                                                                                                                                                                                             
  row = tf.io.parse_example(row, FREEFORM_FEATURES)                                                                                                                                                                      
  row['image'] = tf.reshape(tf.io.decode_raw(row['image'], tf.uint8),
                            [15, 64, 64, 3])                                                                                                  

  row['mask'] = tf.reshape(tf.io.decode_raw(row['mask'], tf.uint8),
                           [15, 64, 64])                                                                                                       
  return row                                                                                                                                                                                                     
                                                                                                                                                                                                                                                                                                                                                                                                                    

def make_tfrecord_paths(dir_name, subdir_name, num_records):
  root = os.path.join(dir_name, subdir_name, 'data.tfrecord')
  paths = [f'{root}-{i:05}-of-{num_records:05}' for i in range(num_records)]
  return paths

def make_freeform_tfrecord_dataset(is_train, shuffle=False, batch_size=32):
  """Returns a TFRecordDataset for freeform data."""
  if is_train:
    subdir_str = 'train'
    num_records = NUM_FREEFORM_TRAIN_TFRECORDS
  else:
    subdir_str = 'test'
    num_records = NUM_FREEFORM_TEST_TFRECORDS

  tfrecord_paths = make_tfrecord_paths('data/freeform', subdir_str, num_records)
  ds = tf.data.TFRecordDataset(tfrecord_paths, compression_type='GZIP')
  ds = ds.map(parse_freeform_row)                                                                                                                                                                 
  return ds


class InputHandle:
    def __init__(self, input_param):
        self.paths = input_param['paths']
        self.num_paths = len(input_param['paths'])
        self.name = input_param['name']
        self.input_data_type = input_param.get('input_data_type', 'float32')
        self.output_data_type = input_param.get('output_data_type', 'float32')
        self.minibatch_size = input_param['minibatch_size']
        self.is_output_sequence = input_param['is_output_sequence']
        self.is_train = input_param['is_train']
        self.data = {}
        self.current_position = 0
        self.load()

    def load(self):
        self.data = make_freeform_tfrecord_dataset(is_train=self.is_train, shuffle=True)
        self.data = self.data.batch(self.minibatch_size)

    def total(self):
        if self.is_train:
            return 300000
        else:
            return 5000

    def begin(self, do_shuffle=True):
        if do_shuffle:
            self.data = self.data.unbatch()
            self.data = self.data.shuffle(buffer_size=50)
            self.data = self.data.batch(self.minibatch_size)
        self.current_position = 0
        self.iterator = self.data.as_numpy_iterator()

    def next(self):
        self.current_position += self.minibatch_size

    def no_batch_left(self):
        if self.current_position >= self.total():
            return True
        else:
            return False

    def get_batch(self):
        curr_data = next(self.iterator)
        return curr_data['image'] / 255.0