import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import random
import os

import logging
logging.getLogger('tensorflow').disabled = True

# title Dataset Constants
NUM_PROBE_TFRECORDS = 20

PROBE_FEATURES = dict(                                                                                                                                                                                                
    possible_image=tf.io.FixedLenFeature(dtype=tf.string, shape=()),
    possible_mask=tf.io.FixedLenFeature(dtype=tf.string, shape=()),
    possible_camera_pose=tf.io.FixedLenFeature(dtype=tf.float32,
                                               shape=(2, 15, 6)),
    impossible_image=tf.io.FixedLenFeature(dtype=tf.string, shape=()),
    impossible_mask=tf.io.FixedLenFeature(dtype=tf.string, shape=()),
    impossible_camera_pose=tf.io.FixedLenFeature(dtype=tf.float32,
                                                 shape=(2, 15, 6)),                                                                                                                            
)                               

# title Dataset Utilities                                                                                                                                                                       
                                                                                                                                                                                                              
def parse_probe_row(row):                                                                                                                                                                                             
  row = tf.io.parse_example(row, PROBE_FEATURES)                                                                                                                                                                      
  for prefix in ['possible', 'impossible']:                                                                                                                                                                      
    row[f'{prefix}_image'] = tf.reshape(
        tf.io.decode_raw(row[f'{prefix}_image'], tf.uint8),
        [2, 15, 64, 64, 3])                                                                                                  
    row[f'{prefix}_mask'] = tf.reshape(
        tf.io.decode_raw(row[f'{prefix}_mask'], tf.uint8),
        [2, 15, 64, 64])                                                                                                       
  return row                                                                                                                                                                                                     

def make_tfrecord_paths(dir_name, subdir_name, num_records):
  root = os.path.join(dir_name, subdir_name, 'data.tfrecord')
  paths = [f'{root}-{i:05}-of-{num_records:05}' for i in range(num_records)]
  return paths

def make_probe_tfrecord_dataset(concept_name, shuffle=False):
  """Returns a TFRecordDataset for probes data."""
  tfrecord_paths = make_tfrecord_paths('data/probes', concept_name, 20)
  ds = tf.data.TFRecordDataset(tfrecord_paths, compression_type='GZIP')
  ds = ds.map(parse_probe_row)                                                                                                                                                                                        
  return ds


class InputHandle:
    def __init__(self, input_param):
        self.paths = input_param['paths']
        self.num_paths = len(input_param['paths'])
        self.name = input_param['name']
        self.input_data_type = input_param.get('input_data_type', 'float32')
        self.output_data_type = input_param.get('output_data_type', 'float32')
        self.minibatch_size = input_param['minibatch_size'] // 4
        self.is_output_sequence = input_param['is_output_sequence']
        self.concept = input_param['concept']
        self.current_position = 0
        self.data = make_probe_tfrecord_dataset(self.concept, shuffle=False).batch(self.minibatch_size)

    def total(self):
        return 5000

    def begin(self, do_shuffle=False):
        self.current_position = 0
        self.iterator = self.data.as_numpy_iterator()

    def next(self):
        self.current_position += self.minibatch_size

    def no_batch_left(self):
        if self.current_position >= self.total():
            return True
        else:
            return False

    def get_batch(self):
        curr_data = next(self.iterator)
        batch_data = np.concatenate([curr_data['possible_image'], curr_data['impossible_image']], axis=1)
        batch_data = batch_data.reshape(-1, 15, 64, 64, 3)
        return batch_data / 255.0