import os.path
import datetime
import cv2 
import numpy as np
import lpips
import torch
from skimage.metrics import structural_similarity
import pickle

from core.utils import preprocess, metrics
from core.utils.density_estimation import KNN, OCSVM, iforest, mahalanobis

loss_fn_alex = lpips.LPIPS(net='alex')

methods = {
    'knn': KNN,
    'ocsvm': OCSVM,
    'iforest': iforest,
    'mahalanobis': mahalanobis,
}


def train(model, ims, real_input_flag, configs, itr):
    cost = model.train(ims, real_input_flag)
    if configs.reverse_input:
        ims_rev = np.flip(ims, axis=1).copy()
        cost += model.train(ims_rev, real_input_flag)
        cost = cost / 2

    if itr % configs.display_interval == 0:
        print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), 'itr: ' + str(itr))
        print('training loss: ' + str(cost))


def test(model, test_input_handle, configs, itr):
    print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), 'test...')
    test_input_handle.begin(do_shuffle=False)
    res_path = os.path.join(configs.gen_frm_dir, str(itr))
    os.mkdir(res_path)
    avg_mse = 0
    batch_id = 0
    img_mse, ssim, psnr = [], [], []
    lp = []

    for i in range(configs.total_length - configs.input_length):
        img_mse.append(0)
        ssim.append(0)
        psnr.append(0)
        lp.append(0)

    # reverse schedule sampling
    if configs.reverse_scheduled_sampling == 1:
        mask_input = 1
    else:
        mask_input = configs.input_length

    real_input_flag = np.zeros(
        (configs.batch_size,
         configs.total_length - mask_input - 1,
         configs.img_width // configs.patch_size,
         configs.img_width // configs.patch_size,
         configs.patch_size ** 2 * configs.img_channel))

    if configs.reverse_scheduled_sampling == 1:
        real_input_flag[:, :configs.input_length - 1, :, :] = 1.0

    while (test_input_handle.no_batch_left() == False):
        batch_id = batch_id + 1
        test_ims = test_input_handle.get_batch()
        test_dat = preprocess.reshape_patch(test_ims, configs.patch_size)
        test_ims = test_ims[:, :, :, :, :configs.img_channel]
        img_gen = model.test(test_dat, real_input_flag)

        img_gen = preprocess.reshape_patch_back(img_gen, configs.patch_size)
        output_length = configs.total_length - configs.input_length 
        img_out = img_gen[:, -output_length:]

        # MSE per frame
        for i in range(output_length):
            x = test_ims[:, i + configs.input_length, :, :, :]
            gx = img_out[:, i, :, :, :]
            gx = np.maximum(gx, 0)
            gx = np.minimum(gx, 1)
            mse = np.square(x - gx).sum()
            img_mse[i] += mse
            avg_mse += mse
            # cal lpips
            img_x = np.zeros([configs.batch_size, 3, configs.img_width, configs.img_width])
            if configs.img_channel == 3:
                img_x[:, 0, :, :] = x[:, :, :, 0]
                img_x[:, 1, :, :] = x[:, :, :, 1]
                img_x[:, 2, :, :] = x[:, :, :, 2]
            else:
                img_x[:, 0, :, :] = x[:, :, :, 0]
                img_x[:, 1, :, :] = x[:, :, :, 0]
                img_x[:, 2, :, :] = x[:, :, :, 0]
            img_x = torch.FloatTensor(img_x)
            img_gx = np.zeros([configs.batch_size, 3, configs.img_width, configs.img_width])
            if configs.img_channel == 3:
                img_gx[:, 0, :, :] = gx[:, :, :, 0]
                img_gx[:, 1, :, :] = gx[:, :, :, 1]
                img_gx[:, 2, :, :] = gx[:, :, :, 2]
            else:
                img_gx[:, 0, :, :] = gx[:, :, :, 0]
                img_gx[:, 1, :, :] = gx[:, :, :, 0]
                img_gx[:, 2, :, :] = gx[:, :, :, 0]
            img_gx = torch.FloatTensor(img_gx)
            lp_loss = loss_fn_alex(img_x, img_gx)
            lp[i] += torch.mean(lp_loss).item()

            real_frm = np.uint8(x * 255)
            pred_frm = np.uint8(gx * 255)

            psnr[i] += metrics.batch_psnr(pred_frm, real_frm)
            for b in range(configs.batch_size):
                score, _ = structural_similarity(pred_frm[b], real_frm[b], full=True, multichannel=True)
                ssim[i] += score

        # save prediction examples
        if batch_id <= configs.num_save_samples:
            path = os.path.join(res_path, str(batch_id))
            os.mkdir(path)
            for i in range(configs.total_length):
                name = 'gt' + str(i + 1) + '.png'
                file_name = os.path.join(path, name)
                img_gt = np.uint8(test_ims[0, i, :, :, :] * 255)
                cv2.imwrite(file_name, img_gt)
            for i in range(output_length):
                name = 'pd' + str(i + 1 + configs.input_length) + '.png'
                file_name = os.path.join(path, name)
                img_pd = img_out[0, i, :, :, :]
                img_pd = np.maximum(img_pd, 0)
                img_pd = np.minimum(img_pd, 1)
                img_pd = np.uint8(img_pd * 255)
                cv2.imwrite(file_name, img_pd)
        test_input_handle.next()

    avg_mse = avg_mse / (batch_id * configs.batch_size)
    print('mse per seq: ' + str(avg_mse))
    for i in range(configs.total_length - configs.input_length):
        print(img_mse[i] / (batch_id * configs.batch_size))

    ssim = np.asarray(ssim, dtype=np.float32) / (configs.batch_size * batch_id)
    print('ssim per frame: ' + str(np.mean(ssim)))
    for i in range(configs.total_length - configs.input_length):
        print(ssim[i])

    psnr = np.asarray(psnr, dtype=np.float32) / batch_id
    print('psnr per frame: ' + str(np.mean(psnr)))
    for i in range(configs.total_length - configs.input_length):
        print(psnr[i])

    lp = np.asarray(lp, dtype=np.float32) / batch_id
    print('lpips per frame: ' + str(np.mean(lp)))
    for i in range(configs.total_length - configs.input_length):
        print(lp[i])


def test_probe(model, test_input_handle, configs, itr):
    # print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), 'test...')
    test_input_handle.begin(do_shuffle=False)
    res_path = os.path.join(configs.gen_frm_dir, str(itr))
    if not os.path.exists(res_path): 
        os.mkdir(res_path)
    batch_id = 0
    tot_surprise = []

    # reverse schedule sampling
    if configs.reverse_scheduled_sampling == 1:
        mask_input = 1
    else:
        mask_input = configs.input_length

    real_input_flag = np.zeros(
        (configs.batch_size,
         configs.total_length - mask_input - 1,
         configs.img_width // configs.patch_size,
         configs.img_width // configs.patch_size,
         configs.patch_size ** 2 * configs.img_channel))

    if configs.reverse_scheduled_sampling == 1:
        real_input_flag[:, :configs.input_length - 1, :, :] = 1.0

    
    # use part of the VoE data to estimate p_VoE
    if configs.voe_estimation:
        hidden_dim = 128
        observation_set = []
        while (test_input_handle.current_position < configs.voe_num_samples / 2):
            batch_id = batch_id + 1
            test_ims = test_input_handle.get_batch()
            test_dat = preprocess.reshape_patch(test_ims, configs.patch_size)
            test_ims = test_ims[:, :, :, :, :configs.img_channel]
            img_gen, features = model.test(test_dat, real_input_flag, return_features=True)
            img_gen = preprocess.reshape_patch_back(img_gen, configs.patch_size)
            output_length = configs.total_length - configs.input_length 
            img_out = img_gen[:, -output_length:]

            features = features.reshape(-1, 4, hidden_dim)
            # extract features of two VoE (physically impossible) videos
            voe_features = features[:, 2:].reshape(-1, hidden_dim)
            observation_set.extend(voe_features)
            test_input_handle.next()
        
        observation_set = np.array(observation_set[:configs.voe_num_samples])
        print("=> successfully collect {} VoE sample to estimate probability density".format(configs.voe_num_samples))
    
    # proposed metric of surprise
    tot_scores = []
    while (test_input_handle.no_batch_left() == False):
        batch_id = batch_id + 1
        test_ims = test_input_handle.get_batch()
        test_dat = preprocess.reshape_patch(test_ims, configs.patch_size)
        test_ims = test_ims[:, :, :, :, :configs.img_channel]
        img_gen, test_features = model.test(test_dat, real_input_flag, return_features=True)
        img_gen = preprocess.reshape_patch_back(img_gen, configs.patch_size)
        output_length = configs.total_length - configs.input_length 
        img_out = img_gen[:, -output_length:]

        # compute surpise for each video clip
        gt = test_ims[:, configs.input_length:configs.input_length+output_length, :, :, :]
        pred = img_out[:, :output_length, :, :, :]
        pred = np.maximum(pred, 0)
        pred = np.minimum(pred, 1)

        # naive surprise: MSE prediction error
        naive_surprise = np.sum(np.mean((pred-gt)**2, axis=(2, 3, 4)), axis=1)

        # subtract the returned value; modified surprise score
        if configs.voe_estimation:
            assert configs.density_estimation in methods.keys()
            gamma = 0.01
            if configs.density_estimation == 'knn':
                k = 50
                scores = methods[configs.density_estimation](observation_set, test_features, k)
            else:
                scores = methods[configs.density_estimation](observation_set, test_features)
            tot_scores.extend(scores)
            surprise = naive_surprise - gamma * scores
        else:
            surprise = naive_surprise

        # compute relative surprise (impossible - possible)
        surprise = surprise.reshape(-1, 4)
        rel_surprise = np.sum(surprise[:, 2:], axis=1) - np.sum(surprise[:, :2], axis=1)
        tot_surprise.extend(rel_surprise)

        # save prediction examples
        '''
        if batch_id <= configs.num_save_samples:
            tot_gt = []
            file_name = os.path.join(res_path, 'gt_{:04d}.png'.format(batch_id))
            for idx in range(4):
                img_gt = []
                border = [255, 255, 255]
                for i in range(configs.total_length):
                    img = np.uint8(test_ims[idx, i, :, :, :] * 255)
                    img = cv2.copyMakeBorder(img, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=border)
                    img_gt.append(img)
                img_gt = np.hstack(img_gt)
                img_gt = cv2.copyMakeBorder(img_gt, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=border)
                tot_gt.append(img_gt)
            tot_gt = np.vstack(tot_gt)
            cv2.imwrite(file_name, tot_gt)
        
            tot_pd = []
            file_name = os.path.join(res_path, 'pred_{:04d}.png'.format(batch_id))
            for idx in range(4):  
                img_pd = []
                for i in range(output_length):
                    img = img_out[idx, i, :, :, :]
                    img = np.maximum(img, 0)
                    img = np.minimum(img, 1)
                    img = np.uint8(img * 255)
                    img = cv2.copyMakeBorder(img, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=border)
                    img_pd.append(img)
                img_pd = np.hstack(img_pd)
                img_pd = cv2.copyMakeBorder(img_pd, 10, 10, 10, 10, cv2.BORDER_CONSTANT, value=border)
                tot_pd.append(img_pd)
            tot_pd = np.vstack(tot_pd)
            cv2.imwrite(file_name, tot_pd)
        '''
        
        test_input_handle.next()

    num_files = len(os.listdir('./score')) 
    file_name = os.path.join('score', '{}.pkl'.format(str(num_files+1)))
    with open(file_name, 'wb') as f:
        pickle.dump(np.array(tot_scores), f)
    
    mean_surprise = np.mean(tot_surprise)
    avg_acc = np.mean(np.array(tot_surprise) > 0)
    print('mean relative surpise: ' + str(mean_surprise))
    print('average accuracy: ' + str(avg_acc))