import numpy as np
import faiss
import sklearn

'''
    to calculate surprise_knn or other surprise that utilize p_VoE estimation,
    just use vanilla surprise substract the return value from the following function.
    estimation includes KNN(), OCSVM(), iforest(), mahalanobis(), 
    their return value are all "distance to VoE distribution"
    You need to prepare observation_set and test_set as feature extracted from model (e.g. the last state of RNN).
'''
def normalize(v):
    '''
    normalize an array to unit norm. The normalization happens at the last axis.
    e.g. [[1,3,5],        [[0.16903085, 0.50709255, 0.84515425],
          [2,4,5],  ->    [0.2981424 , 0.59628479, 0.74535599],
          [3,3,2]]        [0.63960215, 0.63960215, 0.42640143]]
    '''
    v = np.array(v)
    norm = np.linalg.norm(v, axis=-1).reshape(-1,1)
    return v / (norm + 1e-9)

def KNN(
    observation_set: np.array,
    test_set: np.array,
    k: int = 20,
):
    '''
    Input:
        observation_set: an 2D-array, shape = (N, d). The latent representation of VoE samples. N is the number of used VoE samples, d is the feature dimension.
        test_set: an 2D-array, shape = (n, d). The latent representation of test samples. n is the number of test samples, d is the feature dimension.
        k: hyperparameter k for KNN.
    Ouput:
        r_k: an 1D-array, shape=(n). The second term for superise_KNN. 
    '''
    r_k = None

    N, d = observation_set.shape
    n, d = test_set.shape
    
    ## normalize to unit norm
    observation_set = normalize(observation_set)
    test_set = normalize(test_set)

    ## build an index for efficient KNN search
    index = faiss.IndexFlatL2(d)
    index.add(observation_set.astype(np.float32))
    D = index.search(test_set.astype(np.float32), k)

    r_k = D[0][:, -1]

    return r_k

def OCSVM(
    observation_set: np.array,
    test_set: np.array,
):
    from sklearn.svm import OneClassSVM
    clf = OneClassSVM(gamma='auto').fit(observation_set)
    return np.exp(-clf.score_samples(test_set))

def iforest(
    observation_set: np.array,
    test_set: np.array,
):
    from sklearn.ensemble import IsolationForest
    clf = IsolationForest(random_state=0).fit(observation_set)
    return np.exp(-clf.score_samples(test_set))

def mahalanobis(
    observation_set: np.array,
    test_set: np.array,
):
    import sklearn.covariance
    group_lasso = sklearn.covariance.EmpiricalCovariance(assume_centered=False)
    mean = np.mean(observation_set, axis=0)
    X = observation_set - mean
    group_lasso.fit(X)
    precision = group_lasso.precision_
    zero_f = test_set - mean
    term_gau = np.diag(zero_f @ precision @ zero_f.transpose(1,0))
    return term_gau
